; This file is part of the lowest layer of the MINIX kernel.  The other part
; is "proc.c".  The lowest layer does process switching and message handling.
;
; Every transition to the kernel goes through this file.  Transitions are
; caused by sending/receiving messages and by most interrupts.  (RS232
; interrupts may be handled in the file "rs2.x" and then they rarely enter
; the kernel.)
;
; Transitions to the kernel may be nested.  The initial entry may be with a
; system call, exception or hardware interrupt; reentries may only be made
; by hardware interrupts.  The count of reentries is kept in 'k_reenter'.
; It is important for deciding whether to switch to the kernel stack and
; for protecting the message passing code in "proc.c".
; If an interrupt reenters the kernel, it is queued with its parameters r0-r2
; for later processing.
;
; For the message passing trap, most of the machine state is saved in the
; proc table.  (Some of the registers need not be saved.)  Then the stack is
; switched to 'k_stack', and interrupts are reenabled.  Finally, the system
; call handler (in C) is called.  When it returns, interrupts are disabled
; again and the code falls into the restart routine, to finish off held-up
; interrupts and run the process or task whose pointer is in 'proc_ptr'.


; These symbols MUST agree with the values the *.h files to avoid disaster.
; (if they exist)


K_STACK_BYTES	EQU	4096
SAFETY		EQU	512
CLICK_SIZE	EQU	256
IRQ_QUEUE_SIZE	EQU	10
PROC_NR		EQU	27*4


; Offsets of the registers into the proc_table (relative to proc_ptr)
PT_R0		EQU	0
PT_SP		EQU	PT_R0+12*4
PT_SL		EQU	PT_R0+13*4
PT_LK		EQU	PT_R0+14*4
PT_PC		EQU	PT_R0+15*4
PT_R8firq	EQU	PT_PC+4
PT_LKfirq	EQU	PT_R8firq+6*4
PT_SPirq	EQU	PT_LKfirq+4
PT_LKirq	EQU	PT_SPirq+4
PT_SPsvc	EQU	PT_LKirq+4
PT_LKsvc	EQU	PT_SPsvc+4


		GET	$.arm.aslib.register
		GET	$.arm.aslib.arthur
		GET	$.arm.aslib.constants
		GET	$.arm.aslib.macros
		GET	^.^.arm.h.aosdriver
		GET	^.^.arm.h.asconst


;==============================================================================
		AREA	|C$$code|, CODE, READONLY
;==============================================================================

; The external entry points into this file are:

		EXPORT	|__main|	; main entry point
		EXPORT	mpx_2hook	; dummy routine
		EXPORT	idle_task	; executed when there is no work
		EXPORT	clock_int	; clock interrupt routine
		EXPORT	tty_int		; keyboard interrupt routine
		EXPORT	quit_minix	; release all and quit


; This routines are called from within this file

		IMPORT	main
		IMPORT	sys_call
		IMPORT	clock_handler
		IMPORT	keyboard
		IMPORT	lock
		IMPORT	unlock
		IMPORT	unhold
		IMPORT	held_head
		IMPORT	panic
		IMPORT	dis_irq_cnt
		IMPORT	low_user
		IMPORT	high_user
		IMPORT	undef_trap
		IMPORT	pref_trap
		IMPORT	data_trap
		IMPORT	addr_trap
		IMPORT	swap_user_in


;*===========================================================================*
;*				MACROS					     *
;*===========================================================================*
;
; SaveContext: start task switch
;
		MACRO
$label		SaveContext
$label		BL	save
		MEND
;
; RestoreContext: end task switch
;
		MACRO
$label		RestoreContext
$label		B	restart
		MEND


;*===========================================================================*
;*				MINIX					     *
;*===========================================================================*

		ENTRY
MINIX						; this is the entry point
|__main|
		B	MINIX1
		DCD	CLICK_SIZE		; for the boot program

MINIX1		TEQP	PC, #UserMode
		LDR	SP, pk_stktop		; init Stackpointer
		SUB	SL, SP, #K_STACK_BYTES-SAFETY	; set lower end
		LDR	R1, pboot_info
		STR	R0, [R1, #0]		; pointer to boot info block
		MOV	R1, #&7FFFFFFF		; for the moment: no userspace
		LDR	R0, plow_user
		STR	R1, [R0, #0]
		LDR	R0, phigh_user
		STR	R1, [R0, #0]

		MOV	R0, #VecSWI		; This one is to be replaced
		ADR	R1, s_call		; destination address
		SetVec	R0, R1, R2
		LDR	R3, pmnx_swi_vector
		STR	R2, [R3, #0]		; save old vector

		MOV	R0, #VecUndefdInstr	; This one is to be replaced
		ADRL	R1, UndefdInstr		; destination address
		GenBra	R0, R1
		LDR	R2, [R0, #0]
		STR	R1, [R0, #0]
		LDR	R3, pmnx_udef_instr
		STR	R2, [R3, #0]		; save old instruction

		MOV	R0, #VecPrefAbort	; This one is to be replaced
		ADRL	R1, PrefAbort		; destination address
		GenBra	R0, R1
		LDR	R2, [R0, #0]
		STR	R1, [R0, #0]
		LDR	R3, pmnx_pref_instr
		STR	R2, [R3, #0]		; save old instruction

		MOV	R0, #VecDataAbort	; This one is to be replaced
		ADRL	R1, DataAbort		; destination address
		GenBra	R0, R1
		LDR	R2, [R0, #0]
		STR	R1, [R0, #0]
		LDR	R3, pmnx_data_instr
		STR	R2, [R3, #0]		; save old instruction

		MOV	R0, #VecAddrExcept	; This one is to be replaced
		ADRL	R1, AddrExcept		; destination address
		GenBra	R0, R1
		LDR	R2, [R0, #0]
		STR	R1, [R0, #0]
		LDR	R3, pmnx_addr_instr
		STR	R2, [R3, #0]		; save old instr


; this is the point where some initialization is done by the driver:
; e.g. set up the page protection mechanism, set os-mode and more

		SWI	MX_EnableDriver		; init Driver for Minix

		BL	lock
		BL	main			; start the main program
		B	restart			; start scheduler

pk_stktop	DCD	k_stktop
pboot_info	DCD	boot_info
pmnx_swi_vector	DCD	mnx_swi_vector
pmnx_udef_instr	DCD	mnx_udef_instr
pmnx_pref_instr	DCD	mnx_pref_instr
pmnx_data_instr	DCD	mnx_data_instr
pmnx_addr_instr	DCD	mnx_addr_instr
pheld_head	DCD	held_head
pirq_queue_ptr	DCD	irq_queue_ptr
pdis_irq_cnt	DCD	dis_irq_cnt
plow_user	DCD	low_user
phigh_user	DCD	high_user
irq_queue_beg	DCD	irq_queue


;*===========================================================================*
;*				s_call					     *
;*===========================================================================*
s_call						; SWI's are vectored here.
		STMDB	R13!, {R10-R12,LK}	; R13 is SSP!
		BIC	R10, LK, #&FC000003	; clear CC-Bits
		LDR	R12, [R10, #-4]		; fetch SWI-Opcode
		BIC	R11, R12, #&FF000000	; SWI number -> R11
		CMP	R11, #MX_SystemCall	; System Call?
		BEQ	s_call_sys

		SUB	R10, R10, #4		; where does the swi originate?
		LDIND	R12, low_user
		CMP	R10, R12
		BCC	s_call_chksvc
		LDIND	R12, high_user
		CMP	R10, R12
		BCC	s_illeg_swi		; not in userspace ==> error

s_call_chksvc	LDR	R10, =MX_EnterSVC
		CMP	R11, R10
		BNE	s_call_emul

s_enter_svc	LDR	R10, [R13, #3*4]
		ORR	R10, R10, #SVCMode
		STR	R10, [R13, #3*4]
		LDMIA	R13!, {R10-R12,PC}^

s_call_emul	LDMIA	R13!, {R10-R12,LK}	; all registers are restored!
		STMDB	R13!, {R12,PC}
		LDIND	R12, mnx_swi_vector
		STR	R12, [R13, #4]
		LDMIA	R13!, {R12,PC}

s_illeg_swi	LDMIA	R13!, {R10-R12,LK}
		B	UndefdInstr


s_call_sys	LDR	R11, pk_reenter
		LDR	R12, [R11, #0]
		ADD	R12, R12, #1		; increment kernel state
		STR	R12, [R11, #0]

		LDMIA	R13!, {R10,R11}	; leave R12, LK on the stack
		LDIND	R12, proc_ptr	; get address of current proc table
		STMIA	R12!, {R0-R11}	; save most of the registers:R0-R11
		LDMIA	R13!, {R6,R7}	; pop R12->R6 && LK->R7

		STMIA	R12!, {R6}	; save R12
		STMIA	R12!, {R13,LK}^	; save R13_user & R14_user
		STMIA	R12!, {R7}	; save PC
		MOV	R6, R12		; save R12: R12<>R12_firq
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:FIRQMode
		NOP
		STMIA	R6!, {R8-LK}	; save R8_firq 'til R14_firq
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:IRQMode
		NOP
		STMIA	R6!, {R13,LK}	; save R13_irq & R14_irq
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode
		NOP
		STMIA	R6!, {R13,LK}	; save R13_svc & R14_svc

		MOV	R11, R0
		MOV	R0, #1		; enter OS mode
		SWI	MX_OSMode
		MOV	R0, R11

		TEQP	PC, #UserMode	; enable interrupts and enter user mode
		LDR	SP, pk_stktop	; init Stackpointer
		SUB	SL, SP, #K_STACK_BYTES-SAFETY	; set lower end

		MOV	A4, A1		; prepare arguments to sys_call(...)
		MOV	A1, A3		; function (SEND/RECEIVE/BOTH)
		MOV	A3, A2		; pointer to the message
		MOV	A2, A4		; srcdest (MM/FS/...)
		LDIND	V1, proc_ptr	; V1 will be preserved
		BL	sys_call	; process system call
		STR	R0, [V1, #0]	; save returnstatus of sys_call()
		B	restart		; restart process

;s_call_emul	STR	R12, s_dummyswi		; store swi call
;
;		LDR	R12, [R13, #4]		; fetch LK_svc
;		AND	LK, R12, #&FC000003	; preserve flags and mode
;		ADR	R12, s_retfromemul	;
;		ORR	LK, LK, R12		; LK_svc is now complete
;		LDMIA	R13!, {R12}		; restore R12
;		STMDB	R13!, {R12,PC}
;		LDIND	R12, mnx_swi_vector
;		STR	R12, [R13, #4]
;		LDMIA	R13!, {R12,PC}
;
;s_dummyswi	DCD	0			; for original swi routine
;s_retfromemul	STR	PC, s_retmode
;		SWI	MX_EnterSVC
;		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode
;		STMDB	R13!, {R12}		; save register
;
;		LDR	LK, [R13, #4]		; fetch old LK
;		BIC	LK, LK, #&FC000003	; clear flags, preserve address
;		LDR	R12, s_retmode		; fetch new mode & flags
;		AND	R12, R12, #&FC000003	; clear address
;		ORR	LK, LK, R12		; new PC
;		STR	LK, [R13, #4]
;		LDMIA	R13!, {R12, PC}^
;
;s_retmode	DCD	0

;s_illegal_swi	LDMIA	R13!, {R12,PC}^		; NOP!!!



;*===========================================================================*
;*				save					     *
;*===========================================================================*
;
; LK and R0-R3 must be already on the stack already! restart jumps to there...
; Processor mode must be in non-user mode.
; The routine returns in SVC mode, R11 contains new SL
; Level 0 interrupts may be executed but higher levels must be queued
; since then we may have an undefined SP=R12.
; Routine must preserve R0-R2
;
save
		STMDB	R13!, {R0-R2}	; save parameters
		STMDB	R13!, {R11,R12}	; save R11 & R12

		LDR	R11, pk_reenter
		LDR	R12, [R11, #0]
		ADDS	R12, R12, #1		; increment kernel state
		STRLE	R12, [R11, #0]
		BGT	push_current_stack	; stack is already kernels'

		ADD	R0, R13, #5*4	; skip parameters on the stack
		LDMIA	R0, {R0-R3}	; original registers
		LDMIA	R13!, {R11}
		LDIND	R12, proc_ptr	; get address of current proc table
		STMIA	R12!, {R0-R11}	; save most of the registers:R0-R11
		LDMIA	R13!, {R0}	; fetch R12
		STMIA	R12!, {R0}	; save R12
		LDMIA	R13!, {R0-R2}	; fetch parameters again
		ADD	R13, R13, #4*4	; skip already fetched R0-R3
		LDMIA	R13!, {R7}	; fetch return pc
		MOV	R6, LK		; keep return address

		STMIA	R12!, {R13,LK}^	; save R13_user & R14_user
		STMIA	R12!, {R7}	; save PC
		MOV	R5, R12		; R12 <> R12_firq
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:FIRQMode
		NOP
		STMIA	R5!, {R8-LK}	; save R8_firq 'til R14_firq
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:IRQMode
		NOP
		STMIA	R5!, {R13,LK}	; save R13_irq & R14_irq
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode
		NOP
		STMIA	R5!, {R13,LK}	; save R13_svc & R14_svc

		MOV	R3, R0
		MOV	R0, #1		; enter OS mode
		SWI	MX_OSMode
		MOV	R0, R3

		LDR	SP, pk_stktop	; init Stackpointer
		SUB	R11, SP, #K_STACK_BYTES-SAFETY	; set lower end
		MOVS	PC, R6		; return to caller

; now queue the interrupt
push_current_stack
		STMDB	R13!, {R10}
		LDR	R11, pirq_queue_ptr
		LDR	R12, [R11, #0]
		LDR	R10, irq_queue_beg
		SUB	R10, R12, R10
		CMP	R10, #IRQ_QUEUE_SIZE*4*4
		LDMIA	R13!, {R10}
		BLT	pcs_ok
		LDMIA	R13!, {R11,R12}
		ADD	R13, R13, #3*4	; skip parameters
		LDMIA	R13!, {R0-R3,PC}^
pcs_ok
		STMIA	R12!, {R0-R2,LK}; save irq reason and parameters
		STR	R12, [R11, #0]
		LDMIA	R13!, {R11,R12}
		ADD	R13, R13, #3*4	; skip parameters
		LDMIA	R13!, {R0-R3,PC}^


;push_current_stack
;		LDIND	R11, swi_emul_cnt
;		CMP	R11, #0
;		BNV	swi_is_inactive
;
;		LDR	R11, pk_reenter
;		LDR	R12, [R11, #0]
;		SUB	R12, R12, #1	; increment kernel state
;		STR	R12, [R11, #0]
;
;		LDR	R11, pswi_emul_cnt	; increment swi counter
;		LDR	R12, [R11, #0]
;		SUB	R12, R12, #1
;		STR	R12, [R11, #0]
;
;		LDMIA	R13!, {R11,R12,PC}^
;
;		LDR	R11, pirq_queue
;		LDR	R12, [R11, #0]	; increment irq queue pointer
;		ADD	R12, R12, #1
;		STR	R12, [R11, #0]
;		LDR	R11, save_flag
;		STRB	R11, [R12, #-1]	; queue the interrupt
;
;swi_is_inactive
;		LDMIA	R13!, {R11,R12}	; fetch R11 & R12=SP
;		STMDB	SP!, {R0-R11,R13,LK}^	; save R0-R11,R13/14_user
;		LDMIA	R13!, {R0}	; fetch LK (retadr for restart)
;		STMDB	SP!, {R0}	; save LK
;		MOV	R1, LK		; save return address
;		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode
;		NOP
;		STMDB	SP!, {R13,LK}	; save R13_svc, LK_svc
;		LDR	R0, pk_stktop
;		SUB	R0, R0, #K_STACK_BYTES-SAFETY
;		MOVS	PC, R1		; return to caller


;*===========================================================================*
;*				restart 				     *
;*===========================================================================*
restart
		SWI	MX_EnterSVC
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode

		LDIND	R9, irq_queue_ptr
		LDR	R8, irq_queue_beg
		CMP	R9, R8
		BGT	pop_current_stack

		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:UserMode
		LDIND	R0, held_head	; any held interrupts?
		CMP	R0, #0
		BEQ	over_unhold
		BL	lock
		BL	unhold		; flush them!
		BL	unlock
over_unhold
		LDIND	R6, proc_ptr	  ; fetch current process pointer
		LDIND	R7, userprocs	  ; address of userproc table
		CMP	R6, R7		  ; do we restart an user process?
		LDRGE	R0, [R6, #PROC_NR]; yes: fetch process into
		BLGE	swap_user_in	  ; user space
		SWI	MX_EnterSVC
		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode
		CMP	R6, R7		  ; again: is it a user process?
		MOVGE	R0, #0		  ; yes: user processes are running
		SWIGE	MX_OSMode	  ; in user mode! Switch!
no_user_proc
		LDR	R9, pk_reenter
		LDR	R8, [R9, #0]
		SUBS	R8, R8, #1	; decrement kernel state
		STR	R8, [R9, #0]

		LDIND	R0, proc_ptr	  ; fetch current process pointer
		ADD	R0, R0, #2*4	  ; skip R0, R1
		LDMIA	R0!, {R2-R13,LK}^ ; load R2-R12, R13_user, R14_user
		LDMIA	R0!, {R1}	  ; load PC -> R1

;		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:FIRQMode
;		NOP
		ADD	R0, R0, #7*4
;		LDMIA	R0!, {R8-R13,LK}; load R8_firq - R14_firq
;		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:IRQMode
;		NOP
		ADD	R0, R0, #2*4
;		LDMIA	R0!, {R13,LK}	; load R13_irq, R14_irq
;		TEQP	PC, #IRQFlag:OR:FIRQFlag:OR:SVCMode
;		NOP
		ADD	R0, R0, #4	; skip R13_svc
		LDMIA	R0!, {LK}	; fetch LK_svc

		STMDB	R13!, {R1}	; save PC on the stack
		SUB	R1, R0, #PT_LKsvc+4
		LDMIA	R1, {R0,R1}	; load R0, R1
		LDMIA	R13!, {PC}^	; restart process

; unqueue the last interrupt
pop_current_stack
		LDR	R12, pirq_queue_ptr
		LDR	R11, [R12, #0]
		LDMDB	R11!, {R0-R2,LK}
		STR	R11, [R12, #0]
		LDR	SP, pk_stktop
		SUB	R11, SP, #K_STACK_BYTES-SAFETY	; set lower stkpt end
		MOVS	PC, LK

;pop_current_stack
;		LDMIA	SP!, {R13,LK}	; fetch R13_svc & LK_svc
;		LDMIA	SP!, {R0}	; fetch old LK
;		STMDB	R13!, {R0}
;		LDMIA	SP!, {R0-R11,R13,R14}^	; restore registers
;		LDMIA	R13!, {PC}^	; restart process


;*===========================================================================*
;*				clock_int				     *
;*===========================================================================*
clock_int
; "create" a 5Hz interrupt
;		LDR	R0,ccnt
;		ADD	R0, R0, #1
;		CMP	R0, #20
;		MOVGE	R0, #0
;		STR	R0, ccnt
;		LDMLTIA	R13!, {R0-R3,PC}^

		SaveContext
		TEQP	PC, #UserMode
		MOV	SL, R11		; set lower stack end
		BL	clock_handler
		RestoreContext

ccnt		DCD	0


;*===========================================================================*
;*				tty_int 				     *
;*===========================================================================*
tty_int					; Interrupt routine for terminal input
		SaveContext
		TEQP	PC, #UserMode
		MOV	SL, R11
		BL	keyboard	; process a keyboard interrupt
		RestoreContext		; continue execution


;*===========================================================================*
;*				UndefdInstr 				     *
;*===========================================================================*
UndefdInstr
		SUB	LK, LK, #4	; return to the same instruction
		STMDB	R13!, {R0-R3,LK}
		SaveContext
		TEQP	PC, #UserMode
		MOV	SL, R11
		LDIND	V6, proc_ptr	; keep original proc_ptr
		LDR	R0, [V6, #PT_PC]
		BIC	R0, R0, #&FC000003
		BL	undef_trap	; signal process
		LDR	R1, [V6, #PT_PC]
		ADD	R1, R1, R0	; adjust return-PC
		STR	R1, [V6, #PT_PC]
		RestoreContext		; restart another process


;*===========================================================================*
;*				PrefAbort 				     *
;*===========================================================================*
PrefAbort
		SUB	LK, LK, #4	; return to the same instruction
		STMDB	R13!, {R0-R3,LK}
		SaveContext
		TEQP	PC, #UserMode
		MOV	SL, R11
		LDIND	V6, proc_ptr	; keep original proc_ptr
		LDR	R0, [V6, #PT_PC]
		BIC	R0, R0, #&FC000003
		BL	pref_trap
		LDR	R1, [V6, #PT_PC]
		ADD	R1, R1, R0	; adjust return-PC
		STR	R1, [V6, #PT_PC]
		RestoreContext


;*===========================================================================*
;*				DataAbort 				     *
;*===========================================================================*
DataAbort
		SUB	LK, LK, #8	; return to the same instruction
		STMDB	R13!, {R0-R3,LK}
		SaveContext
		TEQP	PC, #UserMode
		MOV	SL, R11
		LDIND	V6, proc_ptr	; keep original proc_ptr
		MOV	R0, V6		; proc_ptr is needed in get_ea
		LDR	R1, [R0, #PT_PC]
		BIC	R1, R1, #&FC000003
		LDR	R1, [R1, #0]	; get faulted instruction
		BL	get_ea		; get effective address in R0
		BL	data_trap
		LDR	R1, [V6, #PT_PC]
		ADD	R1, R1, R0	; adjust return-PC
		STR	R1, [V6, #PT_PC]
		RestoreContext


;*===========================================================================*
;*				AddrExcept 				     *
;*===========================================================================*
AddrExcept
		SUB	LK, LK, #8	; return to the same instruction
		STMDB	R13!, {R0-R3,LK}
		SaveContext
		TEQP	PC, #UserMode
		MOV	SL, R11
		LDIND	V6, proc_ptr	; keep original proc_ptr
		LDR	R0, [V6, #PT_PC]
		BIC	R0, R0, #&FC000003
		BL	addr_trap	; signal process
		LDR	R1, [V6, #PT_PC]
		ADD	R1, R1, R0	; adjust return-PC
		STR	R1, [V6, #PT_PC]
		RestoreContext		; restart another process


;*===========================================================================*
;*				idle_task				     *
;*===========================================================================*
		IMPORT	idle_c
idle_task
; debug
;		SWI	MX_EnterSVC
;		TEQP	PC, #UserMode	; Enable Interrupts

idle_loop	BL	idle_c
;		MOV	R0, #&A00000
;idle_1		SUBS	R0, R0, #1
;		BGE	idle_1
		B	idle_loop	; wait for Interrupt


;*===========================================================================*
;*				mpx_2hook				     *
;*===========================================================================*
mpx_2hook
		MOVS	PC, LK


;*===========================================================================*
;*				get_ea					     *
;*===========================================================================*
get_ea
		AND	R2, R1, #&0C000000	; get type of instruction
		TEQ	R2, #&04000000		; is it a LDR/STR instructon?
		BNE	gea_ldmstm
		AND	R2, R1, #&000F0000
		TEQ	R2, #&000F0000		; is the PC the base register?
		LDR	R2, [R0, R2, LSR #14]	; get base address
		BICEQ	R2, R2, #&FC000003	; clear PSR bits
		SUBEQ	R2, R2, #8		; subtract prefetch
		TST	R1, #&01000000		; postindexing?
		MOVEQ	R0, R2
		MOVEQ	R1, R0
		MOVEQS	PC, LK			; post -> we're ready
		TST	R1, #&02000000		; immediate/register value?
		ANDEQ	R3, R1, #&000000FF	; immediate value is in the
		ANDEQ	R4, R1, #&00000F00	; bits 0 to 11; store them
		ORREQ	R3, R3, R4		; in R3
		BEQ	gea_posneg
		AND	R3, R1, #&0000000F	; get Rm
		LDR	R3, [R0, R3, LSL #2]	; get value of Rm
		AND	R4, R1, #&00000F80	; get shift amount
		MOV	R4, R4, LSR #7
		AND	R5, R1, #&00000060	; get shift type
		MOV	R5, R5, LSR #2		; should be ..., R5 LSR #5
		ADD	PC, PC, R5		; should be ..., R5 LSL #3
		DCD	0
		MOV	R3, R3, LSL R4
		B	gea_posneg
		MOV	R3, R3, LSR R4
		B	gea_posneg
		MOV	R3, R3, ASR R4
		B	gea_posneg
		MOV	R3, R3, ROR R4
gea_posneg	TST	R1, #&00800000		; add/sub the offset?
		SUBEQ	R0, R2, R3		; ea = base - value
		ADDNE	R0, R2, R3		; ea = base + value
		MOV	R1, R0			; address range = 1 word
		MOVS	PC, LK

gea_ldmstm
		AND	R2, R1, #&0E000000
		TEQ	R2, #&08000000		; is it a LDM/STM instruction?
		BNE	gea_panic
		MOV	R2, R1, LSL #16
		MOV	R2, R2, LSR #16
		MOV	R3, #0
gea_count	MOVS	R2, R2, LSR #1		; count number of transferred
		ADC	R3, R3, #0		; registers
		BNE	gea_count
		AND	R2, R1, #&000F0000	; base register
		LDR	R4, [R0, R2, LSR #14]	; get base address
		TST	R1, #&00200000		; writeback specified?
		CMPNE	R2, #&000F0000		; writeback to PC not allowed!
		BEQ	gea_nowback		; no writeback performed
		TST	R1, #&00800000		; up/down?
		ADDEQ	R4, R4, R3, LSL #2	; =0 -> down
		SUBNE	R4, R4, R3, LSL #2	; =1 -> up
		STR	R4, [R0, R2, LSR #14]
gea_nowback	AND	R2, R1, #&01800000	; mask up/down & pre/post
		TST	R1, #&00800000		; up/down
		SUBEQ	R4, R4, R3, LSL #2	; =0 -> down
		TEQ	R2, #0			; decrement after
		TEQNE	R2, #&01800000		; increment before
		ADDEQ	R4, R4, #4
		MOV	R0, R4
		ADD	R1, R0, R3, LSL #2
		SUB	R1, R1, #4
		MOVS	PC, LK

gea_panic
		ADR	R0, gea_panic_msg
		BL	panic
gea_everloop	B	gea_everloop
gea_panic_msg	=	"get_ea: page fault on unknown opcode: ", 0
		ALIGN


;*===========================================================================*
;*				quit_minix				     *
;*===========================================================================*
quit_minix
		LDR	R1, pdis_irq_cnt	; reenable interrupts in the
		MOV	R0, #0			; case they are disabled
		STR	R0, [R1, #0]
		BL	unlock

		SWI	MX_DisableDriver	; restore Minix driver

		MOV	R0, #VecSWI		; restore old vector
		LDIND	R1, mnx_swi_vector
		SetVec	R0, R1, R2

		MOV	R0, #VecUndefdInstr	; restore old vector
		LDIND	R1, mnx_udef_instr
		STR	R1, [R0, #0]

		MOV	R0, #VecPrefAbort	; restore old vector
		LDIND	R1, mnx_pref_instr
		STR	R1, [R0, #0]

		MOV	R0, #VecDataAbort	; restore old vector
		LDIND	R1, mnx_data_instr
		STR	R1, [R0, #0]

		MOV	R0, #VecAddrExcept	; restore old vector
		LDIND	R1, mnx_addr_instr
		STR	R1, [R0, #0]

		SWI	OS_Exit			; sorry, but we must return
						; to Arthur...


;*===========================================================================*
;*===========================================================================*
;*		data for indirect adressing				     *
;*===========================================================================*
;*===========================================================================*

		IMPORT	proc_ptr
		IMPORT	k_reenter
		IMPORT	userprocs

		ALIGN

pproc_ptr	DCD	proc_ptr
pk_reenter	DCD	k_reenter
puserprocs	DCD	userprocs
pint_mess	DCD	0

macbuf		DCD	macbufptr
macbufptr	%	MACBUFSIZE__




;==============================================================================
		AREA	|C$$data|
;==============================================================================

		EXPORT	begdata
|x$dataseg|
begdata
irq_queue_ptr	DCD	irq_queue



;==============================================================================
		AREA	|C$$bss|
;==============================================================================

		EXPORT	sizes
		EXPORT	boot_info
|x$bssseg|
begbss
sizes		%	4*4*4			; the sizes of the four parts
boot_info	%	4
mnx_swi_vector	%	4
mnx_udef_instr	%	4
mnx_pref_instr	%	4
mnx_data_instr	%	4
mnx_addr_instr	%	4
irq_queue	%	IRQ_QUEUE_SIZE*4*4	; 4 words per level
k_stklow	%	K_STACK_BYTES		; descending stack
k_stktop	EQU	.


		END

