/* This file contains the paging mechanism for virtual paging
 * It was written by Martin Patzel
 *
 * This file is concerned with allocating and freeing pages in the physical
 * memory on behalf of the FORK and EXEC system calls.  It also handles
 * allocating and freeing blocks of data on the swap space.  Third, it
 * does the main work concerning swapping pages in and out. The size
 * of a block on the swap device equals to the size of one physical page.
 * The key data structure used is an array of bits each indicating whether
 * the page is free or not. There exists a second array of the same structure
 * for the disk blocks of the swap device.
 * Furthermore there exists three page tables for each process. One for
 * the text-, data- and bss segment. Each page table contains an array
 * of words, one for each page in the virtual address space. Each word
 * contains information about that particular page: whether it's swapped
 * in or out, locked, modified (some attributes may not be supported yet).
 * If the page is in the memory the word contains the physical pagenumber,
 * too.  If the page is swapped out, the word contains the blocknumber on
 * the disk (swap device).
 *
 * The entry points into this file are:
 *   alloc_page:   allocate a physical page
 *   free_page:    release a physical page
 *   mem_init:	initialize the tables when MM start up
 *   max_hole:	returns the largest hole currently available
 *   mem_left:	returns the sum of the sizes of all current holes
 *
 */

#include "kernel.h"
#include <minix/com.h>
#include <minix/callnr.h>
#include <syslib.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include "sconst.h"
#include "proc.h"

#define	DB_ALLOC	0
#define DB_FAULT	0


#define	NO_FREE			(-1)	/* if no free page is found */
#define	NO_MEM			(-1)
#define SAFETY_PAGES		2	/* pages reserved for swapping */


#define	NEXT_MEM_PAGE()		(search_free(mem_map, &mem_map[MEM_MAP_SIZE]))
#define	NEXT_SWAP_PAGE()	(search_free(swap_map, &swap_map[SWAP_MAP_SIZE]))


/* Variables. */
PRIVATE message m;			/* message buffer for in and out */
PRIVATE int result;
PRIVATE int blocksperpage;
PRIVATE int last_who = -1;		/* which user proc is in core? */



_PROTOTYPE(FORWARD do_mem_info, (message *m_ptr)			);
_PROTOTYPE(FORWARD do_ramdisk, (message *m_ptr)				);
_PROTOTYPE(FORWARD do_alloc, (message *m_ptr)				);
_PROTOTYPE(FORWARD do_free, (message *m_ptr)				);
_PROTOTYPE(FORWARD do_fork, (message *m_ptr)				);
_PROTOTYPE(FORWARD do_clrtable, (message *m_ptr)			);
_PROTOTYPE(FORWARD void init_page, (void)				);
_PROTOTYPE(FORWARD phys_pagenum search_free, (bitmap *map, bitmap *endmap));
_PROTOTYPE(FORWARD void build_maps, (int slot)				);
_PROTOTYPE(FORWARD void delete_from_map, (int slot, int vpagenum)	);
_PROTOTYPE(FORWARD phys_pagenum page_to_swap, (void)			);
_PROTOTYPE(FORWARD int alloc_proc_space, (vir_bytes start,
			vir_bytes size, int slot)			);
_PROTOTYPE(FORWARD int free_proc_space, (vir_bytes start,
			vir_bytes size, int slot)			);
#if DB_FAULT
_PROTOTYPE(FORWARD void dump_regs, (struct proc *pr)			);
#endif


/*===========================================================================*
 *				page_task				     * 
 *===========================================================================*/
PUBLIC void page_task()
{
/* Main program of the page task. */

  int r;

  /* First initialize the tables */
  init_page();

  /* Here is the main loop of the page task.  It waits for a message, carries
   * it out, and sends a reply.
   */

  while (TRUE) {
	/* First wait for a request from anywhere */
	receive(ANY, &m);	/* get a request to do some work */

	result = OK;

	/* Now carry out the work. */
  	switch(m.m_type) {
	case PG_MEM_INFO:	r = do_mem_info(&m);	break;
	case PG_RAMDISK:	r = do_ramdisk(&m);	break;
	case PG_ALLOC:		r = do_alloc(&m);	break;
	case PG_FREE:		r = do_free(&m);	break;
	case PG_FORK:		r = do_fork(&m);	break;
	case PG_CLRTABLE:	r = do_clrtable(&m);	break;
	default:		r = EINVAL;		break;
	}

	/* Finally, prepare and send the reply message. */
	m.m_type = r;
	m.m1_i1 = result;
	send(m.m_source, &m);	/* send reply to caller */
  }
}


/*===========================================================================*
 *				do_mem_info				     *
 *===========================================================================*/
PRIVATE do_mem_info(m_ptr)
message *m_ptr;
{
  mem_info *mip = (mem_info *)m_ptr->PAR1;

  mip->low_user    = low_user;
  mip->high_user   = high_user;
  mip->pagesize    = pagesize;
  mip->free_mempg  = free_pages;
  /* some pages reserved for paging */
  mip->free_swappg = free_swap_pages - SAFETY_PAGES;

  return(OK);
}


/*===========================================================================*
 *				do_ramdisk				     *
 *===========================================================================*/
PRIVATE do_ramdisk(m_ptr)
message *m_ptr;
{
  int ram_pages, i;

  ram_pages = (int)(ALIGN_UP(m_ptr->PAR1, pagesize) / pagesize);
  result = (int)(first_page * pagesize);
  for (i = 0; i < ram_pages; i++)
  {
	mem_pagemap[i] = ((long)alloc_page() << 16) |
			 ((long)first_page + (long)i);
  }
  mem_pagemap[i] = -1;
  set_mmu_pagemap(mem_pagemap, PROTECT_OS);
  first_page += ram_pages;

  /* determine how many pages would fit into one PT per process */
  if (last_page >= first_page + MAX_PT_ENTRIES)
	last_page = first_page + MAX_PT_ENTRIES - 1;

  low_user = first_page * pagesize;
  high_user = (last_page + 1L) * pagesize;
  return(OK);
}


/*===========================================================================*
 *				do_alloc				     *
 *===========================================================================*/
PRIVATE do_alloc(m_ptr)
message *m_ptr;
{
  /* allocate memory for a given process */
  return(alloc_proc_space(m_ptr->PAR1, m_ptr->PAR2, m_ptr->PAR3));
}


/*===========================================================================*
 *				do_free					     *
 *===========================================================================*/
PRIVATE do_free(m_ptr)
message *m_ptr;
{
  return(free_proc_space(m_ptr->PAR1, m_ptr->PAR2, m_ptr->PAR3));
}


/*===========================================================================*
 *				do_fork					     *
 *===========================================================================*/
PRIVATE do_fork(m_ptr)
message *m_ptr;
{
  register pt_entry *pppte, *pcpte, *p;
  phys_pagenum pp_num;
  int i, vp_num, parent, child, num;

  parent = m_ptr->PAR1;
  child  = m_ptr->PAR2;

  /* pointer to parent/child page table entries */
  /* let's start at the first page */
  pppte = &pagetables[parent][0];
  pcpte = &pagetables[child][0];
  vp_num = first_page;

  /* look for enough memory */
  for (i = 0, num = 0, p = pppte; i < MAX_PT_ENTRIES; i++, p++)
	if (!(*p & PTE_FREE)) num++;
  if (num > free_pages + free_swap_pages - SAFETY_PAGES)
	return (ENOMEM);

  /* clear pagetable, so that alloc_page can swap correctly */
  for (i = 0, p = pcpte; i < MAX_PT_ENTRIES; i++)
	*p++ = PTE_FREE;
  map_in[child][0] = -1;
  map_out[child][0] = -1;

  for (i = 0, p = pcpte; i < MAX_PT_ENTRIES; i++, p++, pppte++, vp_num++)
  {
  	if (*pppte & PTE_FREE)
  		continue;

	pp_num = alloc_page();
	if (pp_num == NO_FREE)
		panic("alloc: inconsistent free page count", NO_NUM);

	set_mmu_onepage(pp_num, copy_page, PROTECT_FREE);
	if (*pppte & PTE_SWAPPED)
		f_do_devio(READING, Swap_Dev, 0, blocksperpage,
			   ((long)*pppte & PTE_PAGENUM) * (long)blocksperpage,
			   (char *)((long)copy_page * (long)pagesize));
	else
		phys_copy(vp_num * pagesize, copy_page * pagesize, pagesize);
	set_mmu_onepage(pp_num, pool_page, PROTECT_LOCKED);

	*p = pp_num;
  }
  build_maps(child);
  return(OK);
}


/*===========================================================================*
 *				do_clrtable				     *
 *===========================================================================*/
PRIVATE do_clrtable(m_ptr)
message *m_ptr;
{
  int i;
  pt_entry *p = &pagetables[m_ptr->PAR1][0];

  for (i = 0; i < MAX_PT_ENTRIES; i++)
	*p++ = PTE_FREE;
  map_in[m_ptr->PAR1][0] = -1;
  map_out[m_ptr->PAR1][0] = -1;
  return(OK);
}


/*===========================================================================*
 *				alloc_page				     *
 *===========================================================================*/
PUBLIC phys_pagenum alloc_page()
{
/* Allocate a page: First looks for a page in the free page map.
 * If no one is found, a page is being freed (swapped out).
 * This procedure is called when memory is needed for FORK or EXEC.
 */

  phys_pagenum pagenum;

  pagenum = NEXT_MEM_PAGE();
  if (pagenum != NO_FREE)
  {
	if (free_pages-- < 0)
		panic("alloc_page: bad free_pages counter\n", NO_NUM);
#if DB_ALLOC
	printf("ap%d ", pagenum);
#endif
	return(pagenum);
  }

  return(page_to_swap());
}


/*===========================================================================*
 *				free_page				     *
 *===========================================================================*/
PUBLIC void free_page(pagenum)
phys_pagenum pagenum;
{
  if (pagenum < 0 || pagenum > MAX_PHYS_PAGES)
	panic("physical pagenum out of range", NO_NUM);
  SET_BIT(mem_map, pagenum);
  free_pages++;
#if DB_ALLOC
	printf("fp%d ", pagenum);
#endif
}


/*===========================================================================*
 *				free_swap_page				     *
 *===========================================================================*/
PUBLIC void free_swap_page(pagenum)
phys_pagenum pagenum;
{
  if (pagenum < 0 || pagenum >= swap_size)
	panic("swap pagenum out of range", NO_NUM);
  SET_BIT(swap_map, pagenum);
  free_swap_pages++;
#if DB_ALLOC
	printf("fs%d ", pagenum);
#endif
}


/*===========================================================================*
 *				search_free				     *
 *===========================================================================*/
PRIVATE phys_pagenum search_free(map, endmap)
bitmap *map, *endmap;
{
  register bitmap *mapptr, m;
  register phys_pagenum pagenum;

  pagenum = 0;
  for (mapptr = map;
       mapptr < endmap;
       mapptr++, pagenum += sizeof(mapptr) * 8)
  {
	m = *mapptr;
	if (m)
	{
		while ((m & 1) == 0)
			m >>= 1, pagenum++;
		*mapptr &= ~((bitmap)1 << (pagenum % (sizeof(mapptr) * 8)));
		return (pagenum);
	}
  }
  return(NO_FREE);
}


/*===========================================================================*
 *				init_page				     *
 *===========================================================================*/
PRIVATE void init_page()
{
  register bitmap *bmp;
  register int i, j;
  int *drives;
  vir_bytes kernel_start, kernel_end, log_addr;
  struct drive_parm driveparm;
  int mem_freepages[MAX_PHYS_PAGES];	/* list of available physical pages */

  get_mmu_parameters(&mmu_info);	/* get mmu basic parameters */
  get_mmu_pagemap(mem_pagemap);		/* get physical to logical map */
  get_mmu_freepages(mem_freepages);	/* get list of free phys. pages */

  pagesize = mmu_info.mp_actpagesize;
  pool_page = (int)((long)mmu_info.mp_topram / pagesize - 1);
  copy_page = pool_page - 1;
  last_page = copy_page - 1;
  blocksperpage = (int)(pagesize / BLOCK_SIZE);

  kernel_start = ALIGN_DOWN(sizes[SZ_KSTART], pagesize); /* align to page start */
  kernel_end   = ALIGN_UP(sizes[SZ_IEND], pagesize); /* align to next page */
  first_page   = (int)(kernel_end / pagesize);	/* ram disk will adjust it */
  low_user     = first_page * pagesize;
  high_user    = (last_page + 1L) * pagesize;

  swap_size = 0;
  drives = get_drivelist();
  while (*drives >= 0)
  {
  	if ((*drives >> 16) == Swap_Dev)
	{
		get_driveinfo(Swap_Dev, *drives & 0xFFFF, &driveparm);
		swap_size = (int)(driveparm.dp_drvsize / pagesize);
		swap_part = *drives & 0xFFFF;
		break;
	}
	drives++;
  }
  if (swap_size == 0)
	printf("warning: no swap device found\n");

  free_swap_pages = swap_size;
  /* default: all blocks on the swap device are free */
  for (i = 0, bmp = swap_map; i < swap_size; )
  {
	if (i + sizeof(bitmap) * 8 < swap_size)
		*bmp++ = ~(bitmap)0, i += sizeof(bitmap)*8;
	else
		SET_BIT(swap_map, i), i++;
  }

  /* check for unused pages and put them into the page pool */
  free_pages = 0;
  for (i = 0; mem_freepages[i] != -1; i++)
  {
  	log_addr = mem_pagemap[mem_freepages[i]];
  	if (log_addr <  ALIGN_DOWN(kernel_start, pagesize) || 
  	    log_addr >= ALIGN_UP(kernel_end, pagesize))
	{
  		SET_BIT(mem_map, mem_freepages[i]);
		free_pages++;
	}
  }
  /* map all pages in the page pool to the last available page
   * (pool_page) of the free virtual address space
   */
  for (i = 0, j=0; i < mmu_info.mp_actphyspages; i++)
  {
  	if (CHK_BIT(mem_map, i))
  		mem_pagemap[j++] = ((long)i << 16) | pool_page;
  }
  mem_pagemap[j] = -1;
  set_mmu_pagemap(mem_pagemap, PROTECT_LOCKED);
}


/*===========================================================================*
 *				build_maps				     *
 *===========================================================================*/
PRIVATE void build_maps(slot)
int slot;
{
  pt_entry *p;
  phys_pagenum pp_num;
  int i, j;

  p = &pagetables[slot][0];
  for (i = 0, j = 0; i < MAX_PT_ENTRIES; i++, p++)
  {
	if (!(*p & (PTE_FREE | PTE_SWAPPED)))
	{
		pp_num = *p & PTE_PAGENUM;
		map_in[slot][j]  = ((long)pp_num << 16) | ((long)i + first_page);
		map_out[slot][j] = ((long)pp_num << 16) | pool_page;
		j++;
	}
  }
  map_in[slot][j] = -1;
  map_out[slot][j] = -1;
}


/*===========================================================================*
 *				delete_from_map				     *
 *===========================================================================*/
PRIVATE void delete_from_map(slot, vpagenum)
int slot, vpagenum;
{
  int i;
  vir_bytes *vpi, *vpo;

  vpi = &map_in[slot][0];
  for (i = 0; i < MAX_PHYS_PAGES; i++, vpi++)
  {
	if ((*vpi & 0xFFFF) == vpagenum)
		break;
  }
  if (i == MAX_PHYS_PAGES) return;
  vpo = &map_out[slot][i];
  if ((*vpo & 0xFFFF0000) != (*vpi & 0xFFFF0000))
	panic("inconsistent swap maps", NO_NUM);
  for ( ; *vpi != -1; vpi++, vpo++)
  {
	*vpi = *(vpi + 1);
	*vpo = *(vpo + 1);
  }
}


/*===========================================================================*
 *				swap_user_in				     *
 *===========================================================================*/
PUBLIC void swap_user_in(who)
int who;
{
  if (who == last_who || who < INIT_PROC_NR)
	return;

  if (last_who != -1)
  {
	set_mmu_pagemap(map_out[last_who], PROTECT_LOCKED);
	set_mmu_pagemap(map_in[who], PROTECT_FREE);
  }
  last_who = who;
}


/*===========================================================================*
 *				ensure_pages				     *
 *===========================================================================*/
PUBLIC void ensure_pages(slot, start, end)
int slot;
vir_bytes start, end;
{
  pt_entry *ppte, *p;
  phys_pagenum pp_num;
  int start_p, end_p, page, swappage;

  /* calculate in pages */
  start_p = (int)(ALIGN_DOWN(start, pagesize) / pagesize);
  end_p   = (int)(ALIGN_DOWN(end, pagesize) / pagesize);
  ppte = &pagetables[slot][start_p - first_page];

  for (page = start_p, p = ppte; page <= end_p; page++, p++)
  {
	if (*p & PTE_FREE)
		panic("ensure_pages: page is free", NO_NUM);
	if (*p & PTE_SWAPPED)
	{
		swappage = *p & PTE_PAGENUM;
		pp_num = alloc_page();
		if (pp_num == NO_FREE)
			panic("data_trap: no more pages - can't swap", NO_NUM);
		set_mmu_onepage(pp_num, page, PROTECT_FREE);
		f_do_devio(READING, Swap_Dev, 0, blocksperpage,
			   (long)swappage * (long)blocksperpage,
			   (char *)((long)page * (long)pagesize));
		*p = (*p & ~(PTE_SWAPPED | PTE_PAGENUM)) | PTE_LOCKED | pp_num;
		free_swap_page(swappage);
	}
	else
	  	*p |= PTE_LOCKED;
  }

  for (page = start_p, p = ppte; page <= end_p; page++, p++)
	*p &= ~PTE_LOCKED;

  build_maps(slot);
}


/*===========================================================================*
 *				alloc_proc_space			     *
 *===========================================================================*/
PRIVATE int alloc_proc_space(start, size, slot)
vir_bytes start, size;
int slot;
{
  /* allocate memory for a given process */
  pt_entry *ppte, *p;
  phys_pagenum pp_num;
  int i, num;
  int start_p, size_p;

  /* calculate in pages */
  start_p = (int)(ALIGN_DOWN(start, pagesize) / pagesize);
  size_p  = (int)(ALIGN_DOWN(start+size-1, pagesize) / pagesize - start_p + 1);
  ppte = &pagetables[slot][start_p - first_page];

  for (i = 0, num = 0, p = ppte; i < size_p; i++, p++)
	if (*p & PTE_FREE) num++;
  if (num > free_pages + free_swap_pages - SAFETY_PAGES)
	return (ENOMEM);
  for (i = 0, p = ppte; i < size_p; i++, p++)
  {
	if (*p & PTE_FREE)
	{
	  	pp_num = alloc_page();
		if (pp_num == NO_FREE)
			panic("alloc: inconsistent free page count", NO_NUM);
		*p = pp_num;
	}
  }
  build_maps(slot);
  set_mmu_pagemap(map_in[slot], PROTECT_FREE);

  return(OK);
}


/*===========================================================================*
 *				free_proc_space				     *
 *===========================================================================*/
PRIVATE int free_proc_space(start, size, slot)
vir_bytes start, size;
int slot;
{
  register pt_entry *ppte;
  phys_pagenum pp_num;
  int i, j;
  int start_p, size_p;

  start_p = (int)(ALIGN_UP(start-1, pagesize) / pagesize);
  size_p  = (int)(ALIGN_DOWN(start+size, pagesize) / pagesize - start_p);
  if (   start_p < first_page || size_p > MAX_PT_ENTRIES
      || slot < 0 || slot >= NR_PROCS)
	return(E_BAD_SRC);

  /* pointer to page table entries */
  ppte = &pagetables[slot][start_p - first_page];

  for (i = 0, j = 0; i < size_p; i++)
  {
	if (*ppte & PTE_FREE) { ppte++; continue; }
	pp_num = *ppte & PTE_PAGENUM;
	if (*ppte & PTE_SWAPPED)
		free_swap_page(pp_num);
	else
	{
		free_page(pp_num);
		mem_pagemap[j++] = ((long)pp_num << 16) | pool_page;
	}
	*ppte++ = PTE_FREE;
  }
  mem_pagemap[j] = -1;
  set_mmu_pagemap(mem_pagemap, PROTECT_LOCKED);

  build_maps(slot);
  return(OK);
}


/*===========================================================================*
 *				page_to_swap				     *
 *===========================================================================*/
PRIVATE phys_pagenum page_to_swap()
{
  /* Selects a page from the in-core-pages and writes it out to a free
   * place on the swap device. Updates all concerning maps.
   */

  phys_pagenum swappage, ppagenum;
  pt_entry *pt, *pto;
  int vpagenum, found, slot;

  swappage = NEXT_SWAP_PAGE();
  if (free_swap_pages-- < 0)
  {
	if(swappage != NO_FREE)
  		panic("page_to_swap: bad free_swap_pages counter", NO_NUM);
	else
		panic("page_to_swap: no more swap pages available", NO_NUM);
  }
  if (swappage == NO_FREE)
	panic("page_to_swap: no more swap pages available", NO_NUM);

  found = FALSE;
  while (!found)
  {
	/* select a process and a page */
	do
		slot = rand() % (NR_PROCS - INIT_PROC_NR) + INIT_PROC_NR;
	while (proc[slot + NR_TASKS].p_flags & P_SLOT_FREE);
	vpagenum = rand() % MAX_PT_ENTRIES;

	pto = pt = &pagetables[slot][vpagenum];
	do {
		if (*pt & (PTE_FREE | PTE_SWAPPED | PTE_LOCKED)) {
			if (pt-- <= &pagetables[slot][0])
				pt += MAX_PT_ENTRIES;
		}
		else {
			found = TRUE;
			break;
		}
	} while (pt != pto);
  }
  ppagenum = *pt & PTE_PAGENUM;
  vpagenum = (pt - &pagetables[slot][0]) + first_page;
#if DB_ALLOC
  printf("page to be swapped: proc=%d, vpage=%d, ppage=%d, swappage=%d\n",
  	slot, vpagenum, ppagenum, swappage);
#endif
  set_mmu_onepage(ppagenum, copy_page, PROTECT_FREE);
  f_do_devio(WRITING, Swap_Dev, 0, blocksperpage,
	     (long)swappage * (long)blocksperpage,
	     (char *)((long)copy_page * (long)pagesize));
  set_mmu_onepage(ppagenum, pool_page, PROTECT_LOCKED);
  *pt = (*pt & ~PTE_PAGENUM) | PTE_SWAPPED | swappage;
  delete_from_map(slot, vpagenum);
#if DB_ALLOC
	printf("as%d ", ppagenum);
#endif
  return (ppagenum);
}


/*===========================================================================*
 *				addr_trap				     *
 *===========================================================================*/
PUBLIC int addr_trap(addr)
vir_bytes addr;
{
  int slot = proc_number(proc_ptr);

#if DB_FAULT
  printf("address trap caused at address &%08x\n", addr);
#endif
  cause_sig(slot, SIGSEGV);
  addr = addr;
  return(4);
}


/*===========================================================================*
 *				undef_trap				     *
 *===========================================================================*/
PUBLIC int undef_trap(addr)
vir_bytes addr;
{
  int uslot, slot = proc_number(proc_ptr);

  uslot = slot < INIT_PROC_NR ? last_who : slot;
#if DB_FAULT
  printf("undefined instruction trap caused at address &%08x\n", addr);
  printf("caused by proc %d (user-proc %d)\n", slot, uslot);
  dump_regs(proc_addr(uslot));
#endif
  cause_sig(uslot, SIGILL);
  addr = addr;
  return(4);
}


/*===========================================================================*
 *				pref_trap				     *
 *===========================================================================*/
PUBLIC int pref_trap(addr)
vir_bytes addr;
{
  int uslot, slot = proc_number(proc_ptr);
  struct mem_map *dp;
  vir_bytes data_end;

  uslot = slot < INIT_PROC_NR ? last_who : slot;
#if DB_FAULT
  printf("prefetch trap caused at address &%08x\n", addr);
  printf("caused by proc %d at pc=%08x\n", slot, proc_addr(slot)->p_reg.pc);
  printf("caused by user-proc %d\n", uslot);
  dump_regs(proc_addr(uslot));
#endif
  dp = &proc[uslot + NR_TASKS].p_map[D];
  data_end = ((vir_bytes)dp->mem_vir + (vir_bytes)dp->mem_len) << CLICK_SHIFT;

  /* if opcode fetch faulted not in text/data segment, generate SIGSEGV */
  if (addr < low_user || addr >= data_end)
  {
	cause_sig(uslot, SIGSEGV);
	return(4);
  }
  else
	ensure_pages(uslot, addr, addr);

  return(0);
}


/*===========================================================================*
 *				data_trap				     *
 *===========================================================================*/
PUBLIC int data_trap(lo_addr, hi_addr)
vir_bytes lo_addr, hi_addr;
{
  int uslot, slot = proc_number(proc_ptr);
  struct proc *uproc;
  vir_bytes data_end, stackptr;
  int lo_page, hi_page, r;

  uslot = slot < INIT_PROC_NR ? last_who : slot;
  uproc = proc_addr(uslot);
  lo_page = (int)(lo_addr / pagesize - first_page);
  hi_page = (int)(hi_addr / pagesize - first_page);
#if DB_FAULT
  printf("data trap caused between address &%08x and &%08x\n", lo_addr, hi_addr);
  printf("caused by proc %d at pc=%08x\n", slot, proc_addr(slot)->p_reg.pc);
  printf("caused by user-proc %d\n", uslot);
#endif
  data_end = (  (vir_bytes)uproc->p_map[D].mem_vir
  	      + (vir_bytes)uproc->p_map[D].mem_len
  	     ) << CLICK_SHIFT;
  stackptr = uproc->p_reg.sp;

  /* data trap faulted in text/data segment */
  if (lo_addr < data_end && hi_addr < data_end &&
      lo_addr >= low_user && hi_addr >= low_user)
	  ensure_pages(uslot, lo_addr, hi_addr);
  /* data trap faulted in stack segment
   * grant stack extension downwards caused by a push onto the stack
   */
  else if (lo_addr >= stackptr - 4*16 && hi_addr >= stackptr - 4)
  {
	if (pagetables[uslot][lo_addr / pagesize - first_page] & PTE_FREE ||
	    pagetables[uslot][hi_addr / pagesize - first_page] & PTE_FREE)
	{
		r = alloc_proc_space(lo_addr, high_user - lo_addr, uslot);
		if (r != OK)
		{
			cause_sig(uslot, SIGSEGV);
			return(4);
		}
		cause_sig(uslot, SIGSTKFLT);
	}
	else
		ensure_pages(uslot, lo_addr, hi_addr);
  }
  /* data trap faulted in illegal range */
  else
  {
	cause_sig(uslot, SIGSEGV);
	return(4);
  }
  return(0);
}


#if DB_FAULT
PRIVATE void dump_regs(pr)
struct proc *pr;
{
  printf("R0 =%08x, R1 =%08x, R2 =%08x, R3 =%08x\n",
	pr->p_reg.retreg, pr->p_reg.r1, pr->p_reg.r2, pr->p_reg.r3);
  printf("R4 =%08x, R5 =%08x, R6 =%08x, R7 =%08x\n",
	pr->p_reg.r4, pr->p_reg.r5, pr->p_reg.r6, pr->p_reg.r7);
  printf("R8 =%08x, R9 =%08x, R10=%08x, R11=%08x\n",
	pr->p_reg.r8, pr->p_reg.r9, pr->p_reg.r10, pr->p_reg.r11);
  printf("R12=%08x, R13=%08x, R14=%08x, R15=%08x\n",
	pr->p_reg.sp, pr->p_reg.sl, pr->p_reg.lk, pr->p_reg.pc);
}
#endif
